/* This is a PRECC parser for oberon2, based entirely on a Stephen Bevan
 * yacc script. It is not copied, in that I read every line of Stephen's
 * script and then changed everything to PRECC lingo - sort of like
 * translating from french to greek because there aren't exact translations
 * in all cases and I have to duplicate the semantics without the syntax,
 * and sometimes Stephen has had to work around a problem that exists
 * only in yacc so I have been able to write more cleanly. Nevertheless,
 * I have left all the original comments of Stephen's script in and added my
 * own where necessary.
 *
 * Peter T. Breuer 1993 (ptb@comlab.ox.ac.uk)
 *
 * This script (but not the comments) is Copyright P.T. Breuer 1993, but
 * you have my permission to use it as you like.
 *
 * This script uses the same lexer (lex, flex, whatever) as Stephen's so
 * comparisons between PRECC and yacc are possible. Indeed, I have made the
 * measurements.
 */

# define TOKEN int        /* token type gotten from lex  - PTB */

/* Use preccx2?.lib under MSDOS - PTB */

# define BEGIN call_mode=1;if (p_argc>1) freopen(p_argv[1],"rt",stdin);

# include "cc.h"
# include "oberonp2.h"

/* (don't need fancy stuff - PTB) */

/* Stephen says .... */



/* -*- Fundamental -*- */
/* 
** $Id: oberon_2_p.y,v 1.30 1992/06/26 09:07:30 bevan Exp $
*/
/*+file-overview
**
** A YACC/BISON definition for Oberon-2
**
** This is based on the definition given in :-
**
**    The Programming Language Oberon-2
**    H. M\"{o}ssenb\"{o}ck, N. Wirth
**    Institut f\"{u}r Computersysteme, ETH Z\"{u}rich
**    January 1992
**
**-file-overview
**+copyright
**
** Copyright (C) 1991 Department of Computer Science, University of Manchester
** All rights reserved.
**
**-copyright
**+author
**
** Stephen J. Bevan <bevan@cs.man.ac.uk> 19911216
**
**-author
**
** Incompatibilities with real Oberon-2 definition
** ===============================================
**
** Ambiguity problems
** ------------------
**
** Procedure Calls
**
** Oberon-2 is yet another in the long line on languages whose grammar
** is inherently ambiguous.  (This term comes from "Compiler Construction
** for Digital Computers", David Gries, Wiley International, 1971.  I
** haven't found a discussion of this in a more recent compiler book)
**
** An example of this is the procedure call statement.  From the grammar
** as defined in the technical report, the following :-
**
**   a(b)
** 
** could be either a procedure "a" with argument "b" or a procedure "a"
** with a type guard "b".  From the syntax, there is no way to tell them
** apart.  Resolution of this requires extra "semantic" information.
**
** The solution adopted here is to relax the syntax wherever necessary
** and punt the resolution to the "semantic phase".  In the application
** this grammar was written for, this phase is totally separate, but
** the usual solution is to embed some actions directly into the grammar
** to resolve the conflict at parse time, probably by using information
** stored in a symbol table.  If you want this solution, you'll have 
** to write it yourself. 
**
** [ an aside :
**
**  Personally I think that any language that requires symbol table
**  queries to guide parsing is flawed.  I think a better solution
**  is to make the syntax unambiguous.  For example this could
**  be done in Oberon-2 by using a different syntax for the guard
**  type e.g. instead of :-
**
**    a(b)(1,2)
**
**  something like :-
**
**    a<b>(1,2)
**
** ]
**
** Note that the grammar, as defined here, would allow something like :-
**
**   a(b,c)(d,e)
**
** which is totally bogus.  This might seem a bit liberal,
** but the original grammar isn't perfect as it would allow things like :-
**
**   a(b)(c)(d)
**
** which is also bogus.  If you can come up with a "syntax only"
** solution to this, I'd like to hear about it.
**
**
** Qualified Idents
**
** The syntax for referencing a value in another module and for
** referencing a field of a record is the same.  Consequently when a
** fragment such as "a.b" is met, it impossible to decide between the
** two on purely syntactically.  This wouldn't really matter except that
** the ETH grammar attempts to distinguish between module and record 
** references, however it is impossible to make this distinction 
** using a CFG.
**
** The most common way to decide if this is a qualified reference or
** a record designator is to examine the symbol table for "a".  If 
** "a" refers to a module, then "a.b" is a qualified reference, if it is
** a variable, then it is a record reference.
**
** This problem could have been avoided by using different syntax
** for module references.  Possible alternatives are a$b (Clu) or
** a!b (Mesa?)
**
** However, as soon as it is possible to decide which value is required
** based on the access method, it makes it possible to allow the same
** name to be bound to two or more different values.  Wirth's languages
** usually have one single name scope, so this would seem to go against
** the style.  However, Oberon-2 already breaks this rule, by allowing
** overloading of procedure names with resolution based on the receiever
** type.
**
** Still on Qualified Idents ...
**
** In a number of places in the following a Qualified_Ident has been 
** replaced by Type_Reference which is just a synonym for Qualified_Ident.
** The reasons for doing this is that in these cases, the Qualified_Ident
** should be referencing a Type (qualified or not) and as such it makes
** sense to make this intention clear in the syntax (even if it can only
** be enforced using context conditions).
*/

/* ... end of what Stephen said - PTB */

MAIN(Module)

# undef BEGIN
# undef CHAR
# undef END
# undef EXIT


@ ARRAY = <L_ARRAY>

@ BEGIN = <L_BEGIN>

@ BY = <L_BY>

@ CASE = <L_CASE>

@ CONST = <L_CONST>

@ DIV = <L_DIV>

@ DO = <L_DO>

@ ELSE = <L_ELSE>

@ ELSIF = <L_ELSIF>

@ END = <L_END>

@ EXIT = <L_EXIT>

@ FOR = <L_FOR>

@ IMPORT = <L_IMPORT>

@ IF = <L_IF>

@ IN = <L_IN>

@ IS = <L_IS>

@ LOOP = <L_LOOP>

@ MODULE = <L_MODULE>

@ MOD = <L_MOD>

@ NIL = <L_NIL>

@ OF = <L_OF>

@ OR = <L_OR>

@ POINTER = <L_POINTER>

@ PROCEDURE = <L_PROCEDURE>

@ RECORD = <L_RECORD>

@ REPEAT = <L_REPEAT>

@ RETURN = <L_RETURN>

@ THEN = <L_THEN>

@ TO = <L_TO>

@ TYPE = <L_TYPE>

@ UNTIL = <L_UNTIL>

@ VAR = <L_VAR>

@ WHILE = <L_WHILE>

@ WITH = <L_WITH>




@ AMPERSAND = <L_AMPERSAND> 

@ CARET = <L_CARET> 

@ CHAR =  <L_CHAR>

@ CHAR_OR_STRING = <L_CHAR_OR_STRING>

@ CLOSE_BRACE = <L_CLOSE_BRACE> 

@ CLOSE_BRACKET = <L_CLOSE_BRACKET> 

@ CLOSE_PAREN = <L_CLOSE_PAREN> 

@ COLON = <L_COLON> 

@ COLON_EQUALS = <L_COLON_EQUALS>

@ COMMA = <L_COMMA> 

@ EQUALS = <L_EQUALS> 

@ FORWARD_SLASH = <L_FORWARD_SLASH> 

@ HASH = <L_HASH> 

@ HYPHEN = <L_HYPHEN> 

@ IDENTIFIER = <L_IDENTIFIER>

@ INTEGER =  <L_INTEGER>

@ LEFT_CHEVRON = <L_LEFT_CHEVRON> 

@ LEFT_CHEVRON_EQUALS = <L_LEFT_CHEVRON_EQUALS>

@ OPEN_BRACE = <L_OPEN_BRACE> 

@ OPEN_BRACKET = <L_OPEN_BRACKET> 

@ OPEN_PAREN = <L_OPEN_PAREN> 

@ PERIOD = <L_PERIOD> 

@ PERIOD_PERIOD = <L_PERIOD_PERIOD>

@ PLUS = <L_PLUS> 

@ REAL = <L_REAL>

@ RIGHT_CHEVRON = <L_RIGHT_CHEVRON>

@ RIGHT_CHEVRON_EQUALS = <L_RIGHT_CHEVRON_EQUALS>

@ SEMI_COLON = <L_SEMI_COLON> 

@ STAR = <L_STAR> 

@ STRING = <L_STRING>

@ TILDE = <L_TILDE> 

@ VERTICAL_BAR = <L_VERTICAL_BAR> 






@ Add_Op  = Number_Sum
@	 | Number_Difference
@	 | Boolean_Disjunction



@ Assignment
@	 =
@	 Designator
@	 COLON_EQUALS
@	 Expression
@	 


@ Boolean_Disjunction
@	 = OR
@	 



@ Case
@ 	=
@ 	Case_Start
@ 	Expression
@ 	OF
@ 	Case_Items
@ 	Optional_Else	
@ 	Case_End
@ 	


@ Case_End	= END 


@ Case_Item
@ 	= [
@ 	Case_Labels_Sequence
@ 	Case_Item_Separator
@ 	Statement_Sequence
@ 	]


@ Case_Item_Separator = COLON 


@ Case_Items
@ 	= Case_Item { Case_Items_Separator Case_Item }*
@ 	


@ Case_Items_Separator = VERTICAL_BAR 


@ Case_Label	= Constant_Expression 


@ Case_Label_Separator = PERIOD_PERIOD 


@ Case_Labels
@ 	= Case_Label [ Case_Label_Separator Case_Label ]
@ 	


@ Case_Labels_Sequence
@ 	= Case_Labels
@ 	{ Case_Labels_Sequence_Separator Case_Labels }*
@ 	


@ Case_Labels_Sequence_Separator = COMMA 


@ Case_Start	= CASE 




@ Constant_Assignment = EQUALS 


@ Constant_Declaration
@ 	=
@ 	Ident_Def
@ 	Constant_Assignment
@  	Constant_Expression
@ 	Constant_Declaration_Terminator
@ 	


@ Constant_Declarations
@ 	= Constant_Declaration*
@ 	


@ Constant_Declaration_Terminator = SEMI_COLON 



@ Constant_Expression = Expression 


@ Declaration_Simple_Item
@ 	= CONST Constant_Declarations
@ 	| TYPE Type_Declarations
@ 	| VAR Var_Declarations
@ 	


@ Declaration_Simple_Items
@ 	= Declaration_Simple_Item*
@ 	


@ Declaration_Procedure
@ 	= Procedure_Declaration
@ 	| Forward_Declaration
@ 	


@ Declaration_Procedures
@ 	= Declaration_Procedure*
@ 	


@ Declaration_Procedure_Terminator = SEMI_COLON 


@ Declaration_Sequence 
@ 	= Declaration_Simple_Items Declaration_Procedures
@ 	

/* Stephen says ... */

/*
** Designator has been written in a different way from the ETH
** definition to allow the abstract syntaxt tree to be built correctly
** using synthesised attributes.  (The ETH definition is good for
** inherited attributes which are difficult in a L[AL]R system.)
*/

/* End of what Stephen said - PTB */

@ Designator_Base
@ 	=
@ 	Qualified_Ident [ Designator_Record_Ref
@ 	                |
@ 	                Designator_Array_Ref
@ 	                |
@ 	                Designator_Indirection
@ 	                |
@ 	                Designator_Guard
@ 	                ]


@ Designator
@ 	=
@ 	Designator_Base
@ 	           [ Designator_Record_Ref
@ 	           |
@ 	           Designator_Array_Ref
@ 	           |
@ 	           Designator_Indirection
@ 	           |
@ 	           Designator_Guard
@ 	           ]



@ Designator_Array_Ref
@ 	=
@ 	Designator_Array_Ref_Start
@ 	Expression_List
@ 	Designator_Array_Ref_End
@ 	


@ Designator_Array_Ref_End = CLOSE_BRACKET 

@ Designator_Array_Ref_Start = OPEN_BRACKET 

/* Stephen says ... */


/*
** This is where this grammar deviates badly from the ETH definition.
** According to it, only Qualified_Idents should be valid between the
** parens, but this if this is done, it buggers up the 
** procedure/function call syntax.
*/

/* End of what Stephen said - PTB */


@ Designator_Guard
@ 	=
@ 	OPEN_PAREN
@ 	Optional_Expression_List
@ 	CLOSE_PAREN
@ 	

@ Optional_Expression_List
@ 	= [  Expression_List ]
@ 	

@ Designator_Indirection = CARET 



@ Designator_Record_Ref
@ 	=
@ 	PERIOD
@ 	Ident
@ 	




@ Exit	= EXIT 




@ Expression
@     	=
@  	Simple_Expression
@ 	Expression_Optional_Relation_Simple_Expression
@ 	


@ Expression_List
@ 	=  Expression { Expression_List_Separator Expression }*
@ 	


@ Expression_List_Separator = COMMA 


@ Expression_Optional_Relation_Simple_Expression
@ 	= [  Relation Simple_Expression ]
@ 	



@ Factor
@ 	= Factor_Designator
@ 	| Factor_Number
@ 	| Factor_Char
@ 	| Factor_String
@ 	| Factor_Char_Or_String
@ 	| Factor_Nil
@ 	| Factor_Set
@ 	| OPEN_PAREN Expression CLOSE_PAREN 
@ 	| Factor_Negative
@ 	

@ Factor_Designator = Procedure_Call 

@ Factor_Char	= CHAR 

@ Factor_Char_Or_String = CHAR_OR_STRING 

@ Factor_Negative	= Factor_Logical_Negation Factor

@ Factor_Logical_Negation = TILDE 

@ Factor_Nil	= NIL 

@ Factor_Number	= REAL | INTEGER 

@ Factor_Set	= Set 

@ Factor_String	= STRING 




@ For
@ 	=
@ 	For_Start
@ 	Ident
@ 	COLON_EQUALS
@ 	Expression
@ 	TO
@ 	Expression
@ 	For_Optional_By
@ 	DO
@ 	Statement_Sequence
@ 	For_End
@ 	


@ For_End	= END 

@ For_Optional_By
@ 	= [ BY Constant_Expression ]
@ 	


@ For_Start	= FOR 




@ Formal_Param_Section
@ 	=
@ 	Formal_Param_Section_Optional_Var
@ 	Formal_Param_Section_Idents
@ 	Var_Assignment
@ 	Type
@ 	


@ Formal_Param_Section_Optional_Var
@ 	= [  VAR ]
@ 	


@ Formal_Param_Section_Idents
@ 	=
@ 	Ident {
@ 	Formal_Param_Section_Ident_Separator
@ 	Ident }*
@ 	


@ Formal_Param_Section_Ident_Separator = COMMA 



@ Formal_Param_Sections
@ 	= [ Formal_Param_Section_List ]
@ 	


@ Formal_Param_Section_List
@ 	=
@ 	Formal_Param_Section {
@ 	Formal_Param_Section_Separator
@ 	Formal_Param_Section }*
@ 	


@ Formal_Param_Section_Separator = SEMI_COLON 


@ Formal_Params
@ 	=
@ 	Formal_Params_Start
@ 	Formal_Param_Sections
@ 	Formal_Params_End
@ 	Formal_Params_Optional_Type
@ 	


@ Formal_Params_Optional_Type
@ 	= [ COLON Type_Reference ]
@ 	


@ Formal_Params_Start = OPEN_PAREN 


@ Formal_Params_End = CLOSE_PAREN 




@ Forward_Declaration
@ 	=
@ 	Procedure_Declaration_Start
@ 	CARET
@ 	Procedure_Declaration_Optional_Receiver
@ 	Procedure_Declaration_Name
@ 	Procedure_Declaration_Optional_Formal_Params
@ 	Declaration_Procedure_Terminator
@ 		




@ Guard
@ 	=
@ 	Qualified_Ident
@ 	COLON
@ 	Type_Reference
@ 	



@ Ident_Def
@     	=
@ 	Ident [ Ident_Export
@ 	|
@ 	Ident_ReadOnly_Export
@ 	]


@ Ident = IDENTIFIER



@ Ident_List 
@ 	= Ident_Def { Ident_Separator Ident_Def }*
@ 	



@ Ident_Export = STAR 

@ Ident_ReadOnly_Export = HYPHEN 


@ Ident_Separator = COMMA 


/* Stephen says ... */

/* A rather large version of If.
** This has been expanded out somewhat to make it
** easier to construct an abstract syntax tree a
*/

/* End of what Stephen said - PTB */


@ If
@ 	=
@ 	IF
@ 	Expression
@ 	If_Then
@ 	{ If_Else_Ifs
@       |
@       [ If_Else ]
@ 	END
@       }
@ 	

/* that was the recursive way. Now lets be more straightforward - PTB */


@ If_Else
@ 	=
@ 	ELSE
@ 	Statement_Sequence
@ 	

/* Stephen says ... */

/*
** This is right recursive instead of the normal left recursive
** so that the elseif chain will be build up in the correct order.
** A left recursive rule would require reversing the links once the
** list was complete.  Does anybody have a convincing argument why
** this should be written in a left recursive manner and then list
** reversal done? (except for that fact that its easier to write a
** reverse that doesn't have a fixed limit like the YACC stack does)
*/

/* End of what Stephen said - PTB */

@ If_Else_Ifs
@ 	=
@ 	{ ELSIF
@ 	Expression
@ 	If_Then
@       }+
@ 	[ If_Else ]
@       END

/* NB - I didn't see any need for that recursion, so I took it out,
 * but one could try
 *  ELSIF Expression If_Then { If_Else_Ifs | [ If_Else ] END }
 *  - PTB */


@ If_Then
@ 	=
@ 	THEN
@ 	Statement_Sequence
@ 	




@ Import_List
@         = 
@ 	Import_List_Import { Import_List_Separator Import_List_Import }*
       


@ Import_List_Import
@         =
@ 	Ident [ COLON_EQUALS Ident	]
@         


@ Import_List_Optional
@         = [
@ 	IMPORT
@ 	Import_List
@ 	Statement_Separator
@         ]



@ Import_List_Separator = COMMA 




@ Loop
@ 	=
@ 	Loop_Start
@ 	Statement_Sequence
@ 	Loop_End
@ 	


@ Loop_End	= END 


@ Loop_Start	= LOOP 




@ Module
@ 	= 
@ 	Module_Start
@ 	Module_Name_Def
@ 	Statement_Separator
@ 	Import_List_Optional
@ 	Declaration_Sequence
@ 	Module_Optional_Begin_Block
@ 	Module_End_Block
@ 	Module_Name_Def
@         Module_End
@ 	

/* This is my insertion, to control qualified idents more closely */

@ Module_Ident = IDENTIFIER

@ Module_Optional_Begin_Block
@ 	= [ BEGIN Statement_Sequence ]
@ 	


@ Module_Name_Def
@ 	= Ident_Def
@ 	


@ Module_End
@ 	= PERIOD 
@ 	


@ Module_End_Block
@ 	= END
@ 	


@ Module_Start
@ 	= MODULE
@ 	




@ Mul_Op
@ 	= Mul_Op_Product
@ 	| Mul_Op_Real_Quotient
@ 	| Mul_Op_Integer_Quotient
@ 	| Mul_Op_Modulus
@ 	| Mul_Op_Conjunction
@ 	


@ Mul_Op_Product = STAR 


@ Mul_Op_Real_Quotient = FORWARD_SLASH 


@ Mul_Op_Integer_Quotient = DIV 


@ Mul_Op_Modulus = MOD 


@ Mul_Op_Conjunction = AMPERSAND 




@ Number_Sum = PLUS 


@ Number_Difference = HYPHEN 




@ Optional_Else
@ 	= [ ELSE Statement_Sequence ]
@ 	



@ Procedure_Call
@ 	=
@ 	Designator
@ 	




@ Procedure_Declaration
@ 	=
@ 	Procedure_Declaration_Start
@ 	Procedure_Declaration_Optional_Receiver
@ 	Procedure_Declaration_Name
@ 	Procedure_Declaration_Optional_Formal_Params
@ 	Procedure_Declaration_Header_Terminator
@ 	Declaration_Sequence
@ 	Procedure_Declaration_Optional_Begin_Block
@ 	Procedure_Declaration_End_Block
@ 	Procedure_Declaration_Name
@ 	Declaration_Procedure_Terminator
@ 		


@ Procedure_Declaration_End_Block = END 


@ Procedure_Declaration_Header_Terminator = SEMI_COLON 


@ Procedure_Declaration_Optional_Begin_Block
@ 	= [ BEGIN Statement_Sequence ]
@ 	 


@ Procedure_Declaration_Optional_Formal_Params
@ 	= [ Formal_Params ]
@ 	


@ Procedure_Declaration_Optional_Receiver
@ 	= [ Receiver ]
@ 	


@ Procedure_Declaration_Name = Ident_Def 


@ Procedure_Declaration_Start = PROCEDURE 




@ Qualified_Ident
@ 	=
@ 	[ Module_Ident 
@ 	Qualified_Ident_Separator ]
@ 	Ident
@ 	



@ Qualified_Ident_Separator = PERIOD 

@ Qualified_Type_Ident
@ 	=
@ 	[ Module_Ident 
@ 	Qualified_Ident_Separator ]
@ 	Type_Ident
@ 	




@ Receiver
@ 	=
@ 	Receiver_Start
@ 	Receiver_Optional_Var
@ 	Ident
@ 	Var_Assignment
@ 	Ident
@ 	Receiver_End
@ 	


@ Receiver_End = CLOSE_PAREN 



@ Receiver_Optional_Var
@ 	= [ VAR ]
@ 	


@ Receiver_Start = OPEN_PAREN 




@ Relation
@ 	= Relation_Equal
@ 	| Relation_Not_Equal
@ 	| Relation_Less_Than
@ 	| Relation_Less_Equal_To
@ 	| Relation_Greater_Than
@ 	| Relation_Greater_Equal_To
@ 	| Relation_Set_Membership
@ 	| Relation_Type_Test
@ 	

@ Relation_Equal = EQUALS 

@ Relation_Greater_Than = RIGHT_CHEVRON 

@ Relation_Greater_Equal_To = RIGHT_CHEVRON_EQUALS 

@ Relation_Less_Than = LEFT_CHEVRON 

@ Relation_Not_Equal = HASH 

@ Relation_Less_Equal_To = LEFT_CHEVRON_EQUALS 

@ Relation_Set_Membership = IN 

@ Relation_Type_Test = IS 




@ Repeat
@ 	=
@ 	Repeat_Start
@ 	Statement_Sequence
@ 	Repeat_Until
@ 	Expression
@ 	


@ Repeat_Until	= UNTIL 


@ Repeat_Start	= REPEAT 




@ Return	
@ 	=
@ 	RETURN
@ 	Return_Optional_Expression
@ 	


@ Return_Optional_Expression
@ 	= [ Expression ]
@ 	




@ Set
@ 	=
@ 	Set_Start
@ 	[ Set_Element_List ]
@ 	Set_End
@ 	


@ Set_Element
@ 	= Expression [ Set_Element_Separator Expression ]
@ 	


@ Set_Element_List
@ 	= Set_Element { Set_Element_List_Separator Set_Element }*
@ 	

@ Set_Element_List_Separator = COMMA 

@ Set_Element_Separator = PERIOD_PERIOD 

@ Set_End   = CLOSE_BRACE 

@ Set_Start = OPEN_BRACE 



/* this next could be shorter, bt I'll leave it - PTB */

@ Simple_Expression
@ 	=
@ 	Simple_Expression_Optional_Start
@ 	Term [
@ 	Add_Op
@ 	Simple_Expression_Terms
@ 	]


@ Simple_Expression_Terms
@ 	=
@ 	Term { Add_Op Term }*



@ Simple_Expression_Optional_Start
@ 	= [
@         Simple_Expression_Negative 
@ 	| Simple_Expression_Positive 
@       ]


@ Simple_Expression_Negative = HYPHEN 

@ Simple_Expression_Positive = PLUS 



@ Statement
@ 	= Assignment
@ 	| Procedure_Call
@ 	| If
@ 	| Case
@ 	| While
@ 	| Repeat
@ 	| For
@ 	| Loop
@ 	| With
@ 	| Exit
@ 	| Return
@ 	|


@ Statement_Separator = SEMI_COLON 

@ Statement_Sequence 
@ 	=  Statement { Statement_Separator Statement }*
@ 	




@ Term	
@ 	=
@ 	Factor { Mul_Op 
@ 	         Factor }*
@ 	




@ Type 
@ 	= Type_Reference
@ 	| Type_Array
@ 	| Type_Record
@ 	| Type_Pointer
@ 	| Type_Procedure
@ 	

@ Type_Array
@ 	=
@ 	ARRAY
@ 	Type_Array_Bounds
@ 	OF
@ 	Type
@ 	


@ Type_Array_Bounds
@ 	= 
@ 	Constant_Expression { Type_Array_Bounds_Separator Constant_Expression }*
@ 	|


@ Type_Array_Bounds_Separator = COMMA 


@ Type_Assignment = EQUALS 


@ Type_Declaration
@ 	= Ident_Def Type_Assignment Type Type_Declaration_Terminator
@ 	


@ Type_Declarations
@ 	= { Type_Declaration }*
@ 	

@ Type_Declaration_Terminator = SEMI_COLON 

@ Type_Ident = IDENTIFIER

@ Type_Pointer
@ 	=
@ 	POINTER
@ 	TO
@ 	Type
@ 	


@ Type_Procedure
@ 	=
@ 	PROCEDURE
@ 	Procedure_Declaration_Optional_Formal_Params
@ 	

@ 	
@ Type_Record
@ 	=
@ 	RECORD
@ 	Type_Record_Optional_Qualifier
@ 	Type_Record_Field_List
@ 	END
@ 	

@ Type_Record_Field
@ 	= [
@ 	Ident_List
@ 	Type_Record_Field_Assignment
@ 	Type
@ 	]


@ Type_Record_Field_Assignment = COLON 


@ Type_Record_Field_List
@ 	= Type_Record_Field { Type_Record_Field_Separator Type_Record_Field }*
@ 	


@ Type_Record_Field_Separator = SEMI_COLON 


@ Type_Record_Optional_Qualifier
@ 	 = [  Record_Base_Type ]
@  	

@ Record_Base_Type = OPEN_PAREN Type_Reference CLOSE_PAREN

/* Stephen says ... */

/* This is where I find that I have to distinguish type references from
   qualified idents. Idents are distinguished from keywords syntactically
   by containing a lower case letter, but that's too restrictive for their
   use as types, which can be, for example, INTEGER, and don't seem to
   be defined as keywords.

   What I have done is distinguish between qualified type idents and
   qualified (variable) idents, by demanding that the latter are of the form
   [a.]b, whilst the former may be looser and take the form [a.]B . The
   problem only arises when the module ident qualifier isn't there,
   because keywords like END may be mistaken as type idents, but I think
   that a type is always preceded by  a colon.
 */

/* End of what Stephen said - PTB */


@ Type_Reference = Qualified_Type_Ident 


@ Var_Assignment = COLON 


@ Var_Declaration
@ 	= Ident_List Var_Assignment Type Var_Declaration_Terminator
@ 	


@ Var_Declarations
@ 	= { Var_Declaration }*
@ 	


@ Var_Declaration_Terminator = SEMI_COLON 



@ While
@ 	=
@ 	While_Start
@ 	Expression
@ 	DO
@ 	Statement_Sequence
@ 	While_End
@ 	


@ While_End	= END 


@ While_Start	= WHILE 


@ With
@ 	=
@ 	With_Start
@ 	With_Statements
@ 	Optional_Else
@ 	With_End
@ 	


@ With_End	= END 


@ With_Start	= WITH 


@ With_Statement
@ 	=
@ 	Guard
@ 	DO
@ 	Statement_Sequence
@ 	


@ With_Statements
@ 	= With_Statement { With_Statements_Separator With_Statement }*
@ 	


@ With_Statements_Separator = VERTICAL_BAR 

/* End of script - PTB */

